# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2021 JOSECONSCO
# Created by JOSECONSCO

import bpy
from os import path
from bpy_extras import view3d_utils
from mathutils import Matrix, Vector, Quaternion
import gpu
import bmesh
import math
from gpu_extras.batch import batch_for_shader
from .helpers_group import get_bbox_size_and_center, is_inst_coll, unlink_obj_from_scene, link_obj_to_same_collections, matrix_decompose, set_parent, get_obj_bbox_manually
from .array_mod_utils import get_geonodes_mod_type
from .group_pro import GP_OT_GroupProAdd


# get current directory
shader_3d_uniform = gpu.shader.from_builtin('UNIFORM_COLOR') # was 3D_UNIFORM_COLOR
shader_3d_smooth = gpu.shader.from_builtin('SMOOTH_COLOR') # was 3D_UNIFORM_COLOR
use_object = 'Input_11'
inst_input = 'Input_2'
cnt_input = 'Input_3'

LIB_FILE = path.dirname(__file__)+"\\gp_geo_nodes_mods.blend" # new base file - source in hsystem_nodes_src.blend

def import_node_group(node_name, lib_file=LIB_FILE, force_update=False):
    ''' import node gr. from hair_geo_nodes.blend '''
    #TODO: we should also update inner group nodes....
    print(lib_file)
    old_node_gr = bpy.data.node_groups.get(node_name)

    if not old_node_gr or force_update:
        # store name now before it gets overridden
        original_node_groups = {node_group.name: node_group for node_group in bpy.data.node_groups}
        old_node_for_remap = bpy.data.node_groups.get(node_name)
        with bpy.data.libraries.load(lib_file) as (data_from, data_to):
            data_to.node_groups = [node_n for node_n in data_from.node_groups if node_n == node_name]
            imported_gr_names = [node_n for node_n in data_from.node_groups if node_n == node_name]  # stores names
            all_lib_node_groups_names = data_from.node_groups[:] # part gets imported

        if data_to.node_groups:
            print(f"{imported_gr_names=}")
        else:
            print(f"node group {node_name} not found in {lib_file}")
            return None

        if old_node_for_remap: #remap main node
            old_node_for_remap.user_remap(data_to.node_groups[0])
            data_to.node_groups[0].name = node_name
            bpy.data.node_groups.remove(old_node_for_remap)

        return data_to.node_groups[0]
    else:
        return old_node_gr


def move_mouse_to_3d_cursor(context):
    # use context.window.cursor_wrapper to set mouse position - it is using absolute coords (0,0) is bottom left corner of blender window
    cursor_loc = context.scene.cursor.location
    # convert 3d cursor location to screen coords 2d
    screen_coords = view3d_utils.location_3d_to_region_2d(context.region, context.region_data, cursor_loc) # is this relative to region
    # convert region coords to window coords
    screen_coords = (int(screen_coords[0] + context.region.x), int(screen_coords[1] + context.region.y))
    # prefs = bpy.context.preferences
    # offset = prefs.view.ui_scale * 100 # do   x + offset , y + offset - if ui scaled ??
    # set mouse position
    context.window.cursor_warp(*screen_coords)




# write modal operator. On click, it will scene raycast, and calculate distance from active object, to raycasted point
def my_view_orientation(context):
    rv3d = context.space_data.region_3d
    view_inv = rv3d.view_matrix.to_3x3()
    return view_inv.normalized()

def get_orientation_matrix(context):
    orient_type = context.window.scene.transform_orientation_slots[0].type
    mat_x = Vector((1, 0, 0))
    mat_y = Vector((0, 1, 0))
    mat_z = Vector((0, 0, 1))
    custom = context.scene.transform_orientation_slots[0].custom_orientation
    if orient_type == 'CURSOR':
        mat_x = context.scene.cursor.matrix.col[0].xyz
        mat_y = context.scene.cursor.matrix.col[1].xyz
        mat_z = context.scene.cursor.matrix.col[2].xyz
    elif orient_type in {'LOCAL','NORMAL'}:
        obj_mat_w = context.active_object.matrix_world.normalized()
        mat_x = obj_mat_w.col[0].xyz
        mat_y = obj_mat_w.col[1].xyz
        mat_z = obj_mat_w.col[2].xyz
    elif orient_type == 'VIEW':
        view_ma = my_view_orientation(context)
        mat_x = view_ma[0].xyz
        mat_y = view_ma[1].xyz
        mat_z = view_ma[2].xyz
    elif custom:
        mat_x = custom.matrix[0].xyz
        mat_y = custom.matrix[1].xyz
        mat_z = custom.matrix[2].xyz


    orient_mat = Matrix.Identity(3)
    orient_mat.col[0].xyz = mat_x
    orient_mat.col[1].xyz = mat_y
    orient_mat.col[2].xyz = mat_z

    return orient_mat


# function that will create circle mesh object
def get_circle_object(radius=0.1, segments=12):
    # create circle mesh
    circle_obj = bpy.data.objects.get('temp_GP_Circle')
    if not circle_obj:
        mesh = bpy.data.meshes.new('GP_Circle')
        bm = bmesh.new()
        bmesh.ops.create_circle(bm, cap_ends=False, cap_tris=False, radius=radius, segments=segments)
        bm.to_mesh(mesh)
        bm.free()
        mesh.update()
        circle_obj = bpy.data.objects.new('temp_GP_Circle', mesh)
        # circle_obj.scale = (scale, scale, scale)
    return circle_obj


PI2 = 2 * math.pi
def six_circles_coords_calc(points_cnt, center, sub_radius=1.0):
    coords = ()
    for a in range(points_cnt + 1):
        ang_sub = (PI2 * a) / points_cnt
        coords += (center.x + sub_radius*math.cos(ang_sub), center.y + sub_radius*math.sin(ang_sub), center.z), # circle.co + center
    return coords



# draw circle but this time facing circle normal direction
up_vec = Vector((0,0,1))
def normal_circles_coords(points_cnt, center, sub_radius=1.0, circle_normal = Vector((0,0,1))):
    # this time circle normal is not z axis, but circle_normal
    # get any vector perpendicular to circle_normal - will become x axis
    x_axis = circle_normal.cross(up_vec)
    # if x_axis is zero, then circle_normal is z axis, so we can use x axis as y axis - six_circle_coords_calc will work
    if x_axis.length < 0.0001:
        return six_circles_coords_calc(points_cnt, center, sub_radius=sub_radius)
    x_axis.normalize()
    y_axis = circle_normal.cross(x_axis)
    coords = ()
    for a in range(points_cnt + 1):
        ang_sub = (PI2 * a) / points_cnt
        co = center + sub_radius*math.cos(ang_sub)*x_axis + sub_radius*math.sin(ang_sub)*y_axis
        coords += (co.x, co.y, co.z),
    return coords


color = (1.0, 0.5, 0.0, 0.8)
theme = bpy.context.preferences.themes[0]
active_color = (0.2, 1.0, 0.1, 0.8) # theme.view_3d.vertex_select
normal_color = (1.0, 0.7, 0.2, 0.8) #theme.view_3d.vertex

class Shared_GeoNodes_Operator:
    shared_slots = ['pivot_point', 'orientation_matrix', 'state', 'current_location', '_timer', '_text_handle', 'array_holder_obj', 'center', '_handle_draw_verts', 'uniform_matrix', 'circle_ui_scale', 'count', 'init_mod_offset', 'use_external_obj', 'mod_ngroup_name', 'mods_show_viewport']

    @staticmethod
    def draw_header_status(context, text):
        context.area.header_text_set(text)
        context.workspace.status_text_set(text)

    def finish(self, context):
        wm = context.window_manager
        bpy.types.SpaceView3D.draw_handler_remove(self._handle_draw_verts, 'WINDOW')
        context.area.header_text_set(None)
        context.workspace.status_text_set(None)
        context.area.tag_redraw()
        wm.event_timer_remove(self._timer)
        context.scene.tool_settings.use_snap_selectable = False
        self.array_holder_obj.hide_select=False  # XXX:
        self.array_holder_obj.select_set(True)

        # restore drawing of all mods (we disabled mods after current array mods in end_invoke)
        for mod, show_viewport in zip(self.array_holder_obj.modifiers, self.mods_show_viewport):
            mod.show_viewport = show_viewport


    def start_invoke(self, context, event):
        obj = context.active_object
        self.state ='OFFSET' # 'OFFSET' - when adjusting radius, the 'INST_COUNT' - when adjusting array count

        self.orientation_matrix = get_orientation_matrix(context)
        self.uniform_matrix = obj.matrix_world.normalized()
        self.center = obj.matrix_world.translation.copy()

        self._timer = context.window_manager.event_timer_add(0.05, window=context.window)

        context.scene.tool_settings.use_snap_selectable = True # we will disable selection for array_holder, thus it wont snap to itself



    def end_invoke(self, context, event):
        # self.array_holder_obj.select_set(False)
        self.array_holder_obj.hide_select=True  # XXX:
        move_mouse_to_3d_cursor(context)

        # g_rv3d = context.region_data
        # viewDir = g_rv3d.view_rotation @ Vector((0.0, 0.0, 1.0)) # toward camera (-1 away from cam)
        # viewDir.rotate(self.array_holder_obj.matrix_world.inverted()) #  viewDir in array_holder_obj space
        # geo_nodes_mod['Input_9'] = viewDir[:]
        def inst_size(o):
            sizeVec, center = get_bbox_size_and_center(o.instance_collection)
            size = (sizeVec.x + sizeVec.y) * 0.25
            return size


        self.circle_ui_scale = 0
        # check all mods if radial (or linear array are using mod[instance_obj]. if so, use that size
        inst_obj = None
        use_inst_obj = False
        for mod in self.array_holder_obj.modifiers:
            if mod.type == 'NODES':
                if mod.node_group.name == 'GP_RadialArray' or mod.node_group.name == 'GP_LinearArray':
                    use_inst_obj = mod[use_object]
                    inst_obj = mod[inst_input]
                    if use_inst_obj and inst_obj:
                        if is_inst_coll(inst_obj):
                            input_size_name = 'Input_8' if mod.node_group.name == 'GP_RadialArray' else 'Input_9'
                            obj_scale = Vector(mod[input_size_name][:]).length
                            self.circle_ui_scale = inst_size(inst_obj) * obj_scale
                        if self.circle_ui_scale == 0: # try using bbox
                            self.circle_ui_scale = (inst_obj.dimensions.x + inst_obj.dimensions.y) * 0.25
                        if self.circle_ui_scale == 0 and not is_inst_coll(inst_obj): # try using bbox
                            mesh_dim = get_obj_bbox_manually(inst_obj) # world space
                            self.circle_ui_scale = (mesh_dim.x + mesh_dim.y) * 0.25
                    break # first mod found is enough - only it should use inst_obj (if any at all)
        active_obj = context.active_object
        if not use_inst_obj: # use active object / array holder then
            self.circle_ui_scale = (active_obj.dimensions.x + active_obj.dimensions.y) * 0.25
            if self.circle_ui_scale == 0.0: # last ditch effort -- use obj mesh bbox
                mesh_dim = get_obj_bbox_manually(active_obj) # world space
                self.circle_ui_scale = (mesh_dim.x + mesh_dim.y) * 0.25

        if self.circle_ui_scale == 0.0: # last ditch effort
            screen_size_rad_2d = 30 # we want 20px radius in screen space
            # make radius screen_size_rad_2d in screen space, by projecting delta_2d distance to 3d, and calculating radius_3d
            delta_2d = 20
            pt_3d1 = view3d_utils.region_2d_to_location_3d(context.region, context.region_data, (event.mouse_region_x, event.mouse_region_y), self.center)
            pt_3d2 = view3d_utils.region_2d_to_location_3d(context.region, context.region_data, (event.mouse_region_x+delta_2d, event.mouse_region_y), self.center)

            # screen_size_rad_2d / delta_2d = radius_3d / delta_3d
            radius_3d = (pt_3d1-pt_3d2).length * screen_size_rad_2d / delta_2d
            self.circle_ui_scale = radius_3d

        # disable drawing all mods after current one...
        current_idx = 999
        self.mods_show_viewport = []
        for i,mod in enumerate(self.array_holder_obj.modifiers):
            self.mods_show_viewport.append(mod.show_viewport)
            if mod.type == 'NODES':
                if mod.node_group and mod.node_group.name == self.mod_ngroup_name:
                    current_idx = i
            if i > current_idx: # disable drawing all mods after current one...
                mod.show_viewport = False



        context.window_manager.modal_handler_add(self)

        # draw line handler
        args = (self,)
        self._handle_draw_verts = bpy.types.SpaceView3D.draw_handler_add(self._draw_array_ui, args, 'WINDOW', 'POST_VIEW')

        bpy.ops.transform.translate('INVOKE_DEFAULT',  cursor_transform=True,  # release_confirm=True,
                           orient_type='GLOBAL', orient_matrix_type='GLOBAL', orient_matrix=self.orientation_matrix, use_snap_self=False)  # ,snap=event.ctrl



    def pick_best_up_axis(self, context):
        axes =[self.orientation_matrix.col[0].xyz, self.orientation_matrix.col[1].xyz, self.orientation_matrix.col[2].xyz]

        viewDir = context.region_data.view_rotation @ Vector((0.0, 0.0, -1.0))
        # pick best axis based on viewDir dot  up, right, forward
        # use lambda and min to get min dot
        up_axis = max(axes, key=lambda axis: abs(viewDir.dot(axis)))
        # while at it - get ox axis too - it is perpendicular to up_axis
        ox_axis = min(axes, key=lambda axis: abs(up_axis.dot(axis)))
        return up_axis, ox_axis


    def project_cursor_xy(self, context, cursor_loc):
        center_to_cursor = cursor_loc - self.center

        # project cursor to z axis using self.orientation_matrix

        # up_axis = self.orientation_matrix.col[2].xyz
        # ox_axis = self.orientation_matrix.col[0].xyz
        # oy_axis = self.orientation_matrix.col[1].xyz
        # best_matching_up_axis = self.pick_best_up_axis(context)
        best_matching_up_axis =self.best_matching_up_axis # world space..

        # obj_mat_w = context.active_object.matrix_world.normalized()
        # up_axis = obj_mat_w.col[2].xyz
        xy_projected = center_to_cursor - center_to_cursor.project(best_matching_up_axis)
        cursor_loc_projected = self.center + xy_projected
        return cursor_loc_projected, xy_projected


    def raycast_init_3d_cursor(self, context, event, obj):
        scene = context.scene
        region = context.region
        rv3d = context.region_data
        coord_2d = event.mouse_region_x, event.mouse_region_y

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord_2d)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord_2d)

        # ray_target = ray_origin + view_vector
        depsgraph = context.evaluated_depsgraph_get()
        result, hit_loc, normal, index, object, matrix = scene.ray_cast(depsgraph, ray_origin, view_vector)
        if not result:
            depth_location = obj.matrix_world.translation
            hit_loc = view3d_utils.region_2d_to_location_3d(region, rv3d, coord_2d, depth_location)
            self.report({'INFO'}, "No object under cursor")
        return hit_loc


    def add_geo_mod(self, context, obj, node_type = 'GP_RadArray'):
        # create empty mesh object to hold geo nodes modifier
        if self.use_external_obj:
            mesh = bpy.data.meshes.new(node_type)
            array_holder_obj = bpy.data.objects.new('GPro_Array', mesh)
            context.collection.objects.link(array_holder_obj)
            context.view_layer.objects.active = array_holder_obj
            array_holder_obj.matrix_world = obj.matrix_world.copy()
            # set scale to 1, so that geo nodes modifier works properly
            array_holder_obj.scale = (1, 1, 1)
            array_holder_obj.select_set(True)
            obj.select_set(False)

            geo_nodes_mod = array_holder_obj.modifiers.new(node_type,type='NODES')
            node_group = bpy.data.node_groups.get(node_type)
            if not node_group:
                node_group = import_node_group(node_type)

            geo_nodes_mod.node_group = node_group
            # array_holder_obj.modifiers.update()
            # geo_nodes_mod.node_group.update_tag()
            # geo_nodes_mod[inst_obj] = obj

            for area in context.screen.areas:
                if area.type == 'VIEW_3D':
                    area.tag_redraw()

            # unlink obj from scene (since it is now in radial array mod)

            for child in obj.children:
                set_parent(child, array_holder_obj)

            unlink_obj_from_scene(context, obj)
            # check if we are editing group

            GP_OT_GroupProAdd.add_to_group(context, [array_holder_obj])
            return array_holder_obj, geo_nodes_mod
        else:
            geo_nodes_mod = obj.modifiers.new(node_type,type='NODES')
            node_group = bpy.data.node_groups.get(node_type)
            if not node_group:
                node_group = import_node_group(node_type)

            geo_nodes_mod.node_group = node_group
            return obj, geo_nodes_mod



m_2pi = 2 * math.pi

class GP_OT_RadialArrayOperator(bpy.types.Operator, Shared_GeoNodes_Operator):
    bl_idname = "group_pro.radial_array"
    bl_label = "Radial Array"
    bl_description = "Radial Array"
    bl_options = {'REGISTER', 'UNDO'}

    __slots__ = Shared_GeoNodes_Operator.shared_slots + ['best_matching_up_axis','best_x']

    @staticmethod
    def _draw_array_ui(self):
        gpu.state.line_width_set(3)
        gpu.state.blend_set("ALPHA")
        gpu.state.depth_test_set("NONE")

        if self.state == 'OFFSET': # finish after second click
            batch = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": [self.center, self.current_location]})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color",active_color)
            batch.draw(shader_3d_uniform)

            # smooth line from self.current_location to cursor
            batch = batch_for_shader(shader_3d_smooth, 'LINES',
                                     {"pos": (self.current_location, bpy.context.scene.cursor.location),
                                     'color': (active_color, (active_color[0], active_color[1], active_color[2], 0.0))
                                      })
            shader_3d_smooth.bind()
            batch.draw(shader_3d_smooth)


        up_dir = self.best_matching_up_axis

        # draw 6 points on circle (where instances will be placed). in active object space. First point will be placed on self.center + self.current_location
        gpu.state.line_width_set(2)
        gpu.state.blend_set("ALPHA")
        # with gpu.matrix.push_pop():
        #     gpu.matrix.multiply_matrix(self.uniform_matrix)

        # x_axis = self.array_holder_obj.matrix_world.inverted() @ self.current_location
        x_axis = self.current_location - self.center
        x_axis = x_axis - x_axis.project(up_dir) # get x_axis component perpendicular to viewDir
        # quaternion rotation around  viewDir, of x_axis - by 90 degrees
        rot = Quaternion(up_dir, math.radians(90))  # Quat(axis, angle)
        y_axis = x_axis.copy()
        y_axis.rotate(rot)
        # y_axis = Vector((-x_axis[1], x_axis[0], 0.0))
        circle_col = normal_color if self.state == 'OFFSET' else active_color
        for i in range(self.count): # initially we create 6 instances. Create 6 circles for it
            ang_main = (m_2pi * i) / self.count
            sub_circle_center = math.cos(ang_main) * x_axis + math.sin(ang_main) * y_axis
            # sub_circle_coords = six_circles_coords_calc(12, center, sub_radius=self.circle_ui_scale)
            sub_circle_coords = normal_circles_coords(12, self.center+sub_circle_center, sub_radius=self.circle_ui_scale, circle_normal=up_dir)

            batch = batch_for_shader(shader_3d_uniform, 'LINE_LOOP', {"pos": sub_circle_coords})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", circle_col)
            batch.draw(shader_3d_uniform)

        # draw grid 10x10 made form lines. It should be perpendicular to up_dir (use x_axis and y_axis)
        grid_size = 5
        grid_col = (1.0, 1.0, 1.0, 0.1)
        x_axis = self.best_x * self.circle_ui_scale
        y_axis = x_axis.cross(up_dir).normalized() * self.circle_ui_scale
        for i in range(-grid_size, grid_size+1, 1):
            # draw line from -grid_size to grid_size
            batch = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": (self.center + i * x_axis - grid_size * y_axis, self.center + i * x_axis + grid_size * y_axis)})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", grid_col)
            batch.draw(shader_3d_uniform)
            # draw line from -grid_size to grid_size
            batch = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": (self.center + i * y_axis - grid_size * x_axis, self.center + i * y_axis + grid_size * x_axis)})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", grid_col)
            batch.draw(shader_3d_uniform)

        gpu.state.blend_set("NONE")
        gpu.state.line_width_set(1)

        # batch = batch_for_shader(shader_3d_uniform, 'LINE_LOOP', {"pos": circle_co})
        gpu.state.depth_test_set("NONE")
        gpu.state.blend_set("NONE")
        gpu.state.line_width_set(1)


    def offset_accepted(self, context, geo_nodes_mod, reset=False):
        self.state = 'INST_COUNT'
        if self.use_external_obj:
            if reset:
                geo_nodes_mod['Input_7'] = self.init_mod_offset[:]
                self.current_location = self.array_holder_obj.matrix_world @ self.init_mod_offset
        self.draw_header_status(context, f"[MMB Wheel] - Instances Count: {geo_nodes_mod[cnt_input]}   [G] - Radius    [LMB] - Finish")


    def modal(self, context, event):
        geo_nodes_mod = self.array_holder_obj.modifiers.get(self.mod_ngroup_name)
        if event.type == 'TIMER': #update selected objs on timer if any
            if self.state == 'OFFSET':
                context.view_layer.update()  # TO GET UPDATE CURSORLOC

                # NOTE: project 3d cursor, on xy plane of active transformation orientation type.
                cursor_loc = context.scene.cursor.location.copy()
                cursor_loc_projected, xy_projected = self.project_cursor_xy(context, cursor_loc)

                self.current_location = cursor_loc_projected # absolute/world space
                distance = self.current_location.length

                # just update radius
                geo_nodes_mod['Input_7'] = self.array_holder_obj.matrix_world.inverted() @ self.current_location #- self.center
                self.array_holder_obj.update_tag()
                context.area.tag_redraw()

                self.draw_header_status(context, f'[Mouse Drag] - Radius: {distance:.1f}      [LMB] - Accept Radius]')


        # change count of radial array on mmb scroll
        elif event.type == 'WHEELUPMOUSE' and event.ctrl == True:
            if self.state == 'INST_COUNT': # finish after second click
                if geo_nodes_mod:
                    geo_nodes_mod[cnt_input] += 1
                    self.count += 1
                    self.array_holder_obj.update_tag()
                    context.area.tag_redraw()
                    self.draw_header_status(context, f"[Ctrl+MMB Wheel] - Instances Count: {geo_nodes_mod[cnt_input]}   [G] - Radius    [LMB] - Finish")

        elif event.type == 'WHEELDOWNMOUSE' and event.ctrl == True:
            # if self.array_holder_obj:
            if self.state == 'INST_COUNT': # finish after second click
                if geo_nodes_mod:
                    geo_nodes_mod[cnt_input] = max(geo_nodes_mod[cnt_input]-1, 3)
                    self.count = max(self.count-1, 3)
                    self.array_holder_obj.update_tag()
                    context.area.tag_redraw()
                    self.draw_header_status(context, f"[Ctrl+MMB Wheel] - Instances Count: {geo_nodes_mod[cnt_input]}   [G] - Radius    [LMB] - Finish")

        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}        # allow navigation

        elif event.type == 'G' and event.value == 'PRESS' and self.state == 'INST_COUNT': # go back to movement
            self.state = 'OFFSET'
            move_mouse_to_3d_cursor(context)
            bpy.ops.transform.translate('INVOKE_DEFAULT',  cursor_transform=True,  # release_confirm=True,
                                        orient_type='GLOBAL', orient_matrix_type='GLOBAL', orient_matrix=self.orientation_matrix)  # ,snap=event.ctrl

        elif event.type in {'LEFTMOUSE', 'ESC', 'RET'}:
            if self.state == 'OFFSET':
                self.offset_accepted(context, geo_nodes_mod)

            elif self.state == 'INST_COUNT': # finish after second click
                self.finish(context)
                return {'FINISHED'}

        elif event.type == 'RIGHTMOUSE' and event.value == 'RELEASE':
            if self.state == 'OFFSET':
                self.offset_accepted(context, geo_nodes_mod, reset=True)

            elif self.state == 'INST_COUNT':
                self.finish(context)
                bpy.ops.ed.undo() # properly undo
                return {'CANCELLED'}

        # return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}


    def invoke(self, context, event):
        obj = context.active_object
        if obj is None:
            self.report({'ERROR'}, "No active object selected. C")
            return {'CANCELLED'}

        self.start_invoke(context, event)
        self.draw_header_status(context, f'[Mouse Drag] - Radius: {0.0:.1f}      [LMB] - Confirm]')
        self.mod_ngroup_name = 'GP_RadialArray'
        geo_nodes_mod = obj.modifiers.get(self.mod_ngroup_name)
        if geo_nodes_mod: # old mod already exists
            self.use_external_obj = geo_nodes_mod[use_object]
            if self.use_external_obj and not geo_nodes_mod[inst_input]: # use extra obj
                self.use_external_obj = False # needed (if inst_obj is missing but geo_nodes_mod[use_object] is True)  (user error?)

            self.array_holder_obj = obj
            self.count = geo_nodes_mod[cnt_input]
            self.init_mod_offset = Vector(geo_nodes_mod['Input_7'][:])

            self.current_location = self.array_holder_obj.matrix_world @ self.init_mod_offset
            context.scene.cursor.location = self.current_location.copy()
            # context.view_layer.update()  # TO GET UPDATE CURSORLOC

            local_up = Vector(geo_nodes_mod['Input_9'])
            local_up.rotate(self.array_holder_obj.matrix_world) # local_up in world space
            self.best_matching_up_axis = local_up.copy()  # now in world space
            # how do we get best x axis? Use init_mod_offset normalized
            best_x = self.init_mod_offset.normalized()
            best_x.rotate(self.array_holder_obj.matrix_world) # best_x in world space
            self.best_x = best_x

        else:
            self.use_external_obj = is_inst_coll(obj)
            self.count = 6 # default value
            self.array_holder_obj, geo_nodes_mod = self.add_geo_mod(context, obj, node_type=self.mod_ngroup_name)

            orient_type = context.window.scene.transform_orientation_slots[0].type
            if orient_type != 'LOCAL': # for local orientation, we will keep array holder orientation
                geo_nodes_mod['Input_8'] = obj.scale[:] # set scale to obj scale
                geo_nodes_mod['Input_10'] = obj.rotation_euler[:] # set scale to obj scale

                depsgraph = context.evaluated_depsgraph_get()  # write previous mw into obj
                self.array_holder_obj.matrix_world = Matrix.Identity(4)
                self.array_holder_obj.matrix_world.translation = self.center


            if self.use_external_obj:

                geo_nodes_mod[use_object] = True # use extra object: None -> to hide instanced objs
                geo_nodes_mod[inst_input] = obj
            else:
                geo_nodes_mod[use_object] = False # use extra object: None -> to hide instanced objs
                geo_nodes_mod[inst_input] = None

            geo_nodes_mod[cnt_input] = 6     # count
            self.init_mod_offset = Vector((1,0,0))
            self.array_holder_obj.update_tag()

            self.current_location = self.array_holder_obj.matrix_world @ Vector((0,0,0))
            context.scene.cursor.location = self.current_location.copy()


            self.best_matching_up_axis, self.best_x = self.pick_best_up_axis(context)
            best_matching_up_axis_local = self.best_matching_up_axis.copy()
            best_matching_up_axis_local.rotate(self.array_holder_obj.matrix_world.inverted()) # best_matching_up_axis in array_holder_obj space
            geo_nodes_mod['Input_9'] = best_matching_up_axis_local[:]


        self.end_invoke(context, event)  # start timer, move mouse to 3d cursor, draw handlers , etc
        return {'RUNNING_MODAL'}



class GP_OT_LinearArrayOperator(bpy.types.Operator, Shared_GeoNodes_Operator):
    bl_idname = "group_pro.linear_array"
    bl_label = "Linear Array"
    bl_description = "Linear Array"
    bl_options = {'REGISTER', 'UNDO'}

    __slots__ = Shared_GeoNodes_Operator.shared_slots + ['b_offset_mode','view_projection_matrix']

    @staticmethod
    def _draw_array_ui(self):
        gpu.state.line_width_set(3)
        gpu.state.blend_set("ALPHA")
        gpu.state.depth_test_set("NONE")

        if self.state == 'OFFSET': # finish after second click
            batch = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": [self.center, self.current_location]})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color",active_color)
            batch.draw(shader_3d_uniform)

            # smooth line from self.current_location to cursor
            batch = batch_for_shader(shader_3d_smooth, 'LINES',
                                     {"pos": (self.current_location, bpy.context.scene.cursor.location),
                                     'color': (active_color, (active_color[0], active_color[1], active_color[2], 0.0))
                                      })
            shader_3d_smooth.bind()
            batch.draw(shader_3d_smooth)


        # draw 6 points on circle (where instances will be placed). in active object space. First point will be placed on self.center + self.current_location
        gpu.state.line_width_set(2)
        gpu.state.blend_set("ALPHA")

        context = bpy.context
        g_rv3d = context.region_data
        # view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        # ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
        viewDir = g_rv3d.view_rotation @ Vector((0.0, 0.0, -1.0))

        # x_axis = self.current_location - self.center
        # x_axis = self.array_holder_obj.matrix_world.inverted() @ self.current_location
        x_axis = self.current_location - self.center
        # do lerp from 0 to x_axis - and draw self.count circles along this line
        circle_col = normal_color if self.state == 'OFFSET' else active_color
        for i in range(self.count): # initially we create 6 instances. Create 6 circles for it
            if self.b_offset_mode:
                center = x_axis * i + self.center
            else:
                center = x_axis * i/(self.count-1) + self.center
            # sub_circle_coords = six_circles_coords_calc(12, center, sub_radius=self.circle_ui_scale)
            sub_circle_coords = normal_circles_coords(12, center, sub_radius=self.circle_ui_scale, circle_normal=viewDir)

            batch = batch_for_shader(shader_3d_uniform, 'LINE_LOOP', {"pos": sub_circle_coords})
            shader_3d_uniform.bind()
            shader_3d_uniform.uniform_float("color", circle_col)
            batch.draw(shader_3d_uniform)

        gpu.state.blend_set("NONE")
        gpu.state.line_width_set(1)

        # batch = batch_for_shader(shader_3d_uniform, 'LINE_LOOP', {"pos": circle_co})
        gpu.state.depth_test_set("NONE")
        gpu.state.blend_set("NONE")
        gpu.state.line_width_set(1)


    def offset_accepted(self, context, geo_nodes_mod, reset=False):
        self.state = 'INST_COUNT'
        if self.use_external_obj:
            if reset: # must have been called from reset (RMB)
                geo_nodes_mod['Input_7'] = self.init_mod_offset # reset offset
                self.current_location = self.array_holder_obj.matrix_world @ Vector(self.init_mod_offset) # and current location
        self.draw_header_status(context, f"[MMB Wheel] - Instances Count: {geo_nodes_mod[cnt_input]}    [X] - Offset Mode   [G] - Move Instances   [LMB] - Finish")


    def modal(self, context, event):
        geo_nodes_mod = self.array_holder_obj.modifiers.get(self.mod_ngroup_name)
        if event.type == 'TIMER': #update selected objs on timer if any
            if self.state == 'OFFSET':
                context.view_layer.update()  # TO GET UPDATE CURSORLOC

                self.current_location = context.scene.cursor.location.copy() # absolute..
                distance = self.current_location.length

                # just update radius
                if geo_nodes_mod:
                    geo_nodes_mod['Input_7'] = self.array_holder_obj.matrix_world.inverted() @ self.current_location #- self.center
                self.array_holder_obj.update_tag()
                context.area.tag_redraw()

                self.draw_header_status(context, f'[Mouse Drag] - Spacing: {distance:.1f}      [LMB] - Accept Spacing]')

        # change count of radial array on mmb scroll
        elif event.type == 'WHEELUPMOUSE' and event.ctrl == True:
            if self.state == 'INST_COUNT': # finish after second click
                if geo_nodes_mod:
                    geo_nodes_mod[cnt_input] += 1
                    self.count += 1
                    self.array_holder_obj.update_tag()
                    context.area.tag_redraw()
                    self.draw_header_status(context, f"[Ctrl+MMB Wheel] - Instances Count: {geo_nodes_mod[cnt_input]}    [X] - Offset Mode   [G] - Move Instances   [LMB] - Finish")

        elif event.type == 'WHEELDOWNMOUSE' and event.ctrl == True:
            # if self.array_holder_obj:
            if self.state == 'INST_COUNT': # finish after second click
                if geo_nodes_mod:
                    geo_nodes_mod[cnt_input] = max(geo_nodes_mod[cnt_input]-1, 3)
                    self.count = max(self.count-1, 3)
                    self.array_holder_obj.update_tag()
                    context.area.tag_redraw()
                    self.draw_header_status(context, f"[Ctrl+MMB Wheel] - Instances Count: {geo_nodes_mod[cnt_input]}    [X] - Offset Mode   [G] - Move Instances   [LMB] - Finish")

        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}        # allow navigation


        elif event.type == 'G' and event.value == 'PRESS' and self.state == 'INST_COUNT': # go back to movement
            self.state = 'OFFSET'
            move_mouse_to_3d_cursor(context)
            bpy.ops.transform.translate('INVOKE_DEFAULT',  cursor_transform=True,  # release_confirm=True,
                                        orient_type='GLOBAL', orient_matrix_type='GLOBAL', orient_matrix=self.orientation_matrix)  # ,snap=event.ctrl

        elif event.type == 'X' and event.value == 'PRESS':
            if self.state == 'INST_COUNT':
                geo_nodes_mod['Input_8'] = not geo_nodes_mod['Input_8'] # switch offset mode
                self.b_offset_mode = geo_nodes_mod['Input_8']
                self.array_holder_obj.update_tag()
                context.area.tag_redraw()

        elif event.type in {'LEFTMOUSE', 'ESC', 'RET'}:
            if self.state == 'OFFSET':
                self.offset_accepted(context, geo_nodes_mod, reset=False)

            elif self.state == 'INST_COUNT': # finish after second click
                self.finish(context)
                return {'FINISHED'}

        elif event.type == 'RIGHTMOUSE' and event.value == 'RELEASE':
            if self.state == 'OFFSET':
                self.offset_accepted(context, geo_nodes_mod, reset=True)

            elif self.state == 'INST_COUNT':
                self.finish(context)
                bpy.ops.ed.undo() # properly undo
                return {'CANCELLED'}

        # return {'PASS_THROUGH'}
        return {'RUNNING_MODAL'}


    def invoke(self, context, event):
        obj = context.active_object
        if obj is None:
            self.report({'ERROR'}, "No active object selected. C")
            return {'CANCELLED'}

        self.start_invoke(context, event)

        self.draw_header_status(context, f'[Mouse Drag] - Spacing: {0.0:.1f}      [LMB] - Confirm]')
        self.mod_ngroup_name = 'GP_LinearArray'

        geo_nodes_mod = obj.modifiers.get(self.mod_ngroup_name)
        if geo_nodes_mod:
            # Disable drawn instance mes while we are in 'DRAG' state
            self.use_external_obj = geo_nodes_mod[use_object]
            if self.use_external_obj and not geo_nodes_mod[inst_input]: # use extra obj
                self.use_external_obj = False # needed (if inst_obj is missing but geo_nodes_mod[use_object] is True)  (user error?)

            self.array_holder_obj = obj
            self.init_mod_offset = Vector(geo_nodes_mod['Input_7'][:])

            # context.scene.cursor.location = self.raycast_init_3d_cursor(context, event, obj)
            self.current_location = self.array_holder_obj.matrix_world @ self.init_mod_offset
            context.scene.cursor.location = self.current_location.copy()

        else:
            self.use_external_obj = is_inst_coll(obj)
            # get children of obj
            self.array_holder_obj, geo_nodes_mod = self.add_geo_mod(context, obj, node_type=self.mod_ngroup_name)

            orient_type = context.window.scene.transform_orientation_slots[0].type
            if orient_type != 'LOCAL': # for local orientation, we will keep array holder orientation
                geo_nodes_mod['Input_9'] = obj.scale[:] # set scale to obj scale
                geo_nodes_mod['Input_10'] = obj.rotation_euler[:] # set scale to obj scale

                depsgraph = context.evaluated_depsgraph_get()  # write previous mw into obj
                self.array_holder_obj.matrix_world = Matrix.Identity(4)
                self.array_holder_obj.matrix_world.translation = self.center

            if self.use_external_obj:
                geo_nodes_mod[use_object] = True # use extra object: None -> to hide instanced objs
                geo_nodes_mod[inst_input] = obj
            else:
                geo_nodes_mod[use_object] = False # use extra object: None -> to hide instanced objs
                geo_nodes_mod[inst_input] = None

            geo_nodes_mod[cnt_input] = 5     # count

            self.array_holder_obj.update_tag()
            self.init_mod_offset = Vector((1,0,0))

            # context.scene.cursor.location = self.raycast_init_3d_cursor(context, event, obj)
            self.current_location = self.center.copy()
            context.scene.cursor.location = self.center.copy()

        self.count = geo_nodes_mod[cnt_input]
        self.b_offset_mode = geo_nodes_mod['Input_8']


        self.end_invoke(context, event) # start timer, move mouse to 3d cursor, draw handlers etc
        return {'RUNNING_MODAL'}



# dissolve radial array mod - recreates Geometry nodes in python
# use circle normal axis ['Input_9'], circle instances count input [cnt_input], and circle OX axis ['Input_7']
# TODO: add scale support...

class GPRO_OT_DissolveRadialArray(bpy.types.Operator):
    bl_label = "Dissolve Radial Array"
    bl_idname = "group_pro.dissolve_radial_array"
    bl_description = "Dissolve Radial Array"
    bl_options = {'UNDO', 'REGISTER', 'INTERNAL'}

    @staticmethod
    def dissolve_radial_array_mod(context, array_holder_obj):
        array_mod = get_geonodes_mod_type(array_holder_obj, 'GP_RadialArray')
        if not array_mod:
            print("No Radial Array modifier found")
            return False
        # get inputs
        circle_normal = Vector(array_mod['Input_9'])
        circle_count = array_mod[cnt_input]
        circle_ox = Vector(array_mod['Input_7'])
        instanced_obj = array_mod[inst_input]

        # decompose instanced_obj matrix_world - we will use its rotation compoenet later
        obj_mat_loc, obj_mat_rot, obj_mat_sca = matrix_decompose(instanced_obj.matrix_world)


        # get ox component perpendicular to circle_normal
        circle_ox_perp = circle_ox - circle_ox.project(circle_normal)

        # oy compoenent (ox rotated 90 around circle_normal)
        rot_90 = Quaternion(circle_normal, math.radians(90))

        circle_oy = circle_ox_perp.copy()
        circle_oy.rotate(rot_90)

        circle_center = circle_ox.project(circle_normal)


        new_objs = []
        # create circle
        for i in range(circle_count):
            angle = i * 2 * math.pi / circle_count
            circle_pos = circle_center + circle_ox_perp * math.cos(angle) + circle_oy * math.sin(angle)

            # instance rotation = quat(circle_normal, angle)
            rot = Quaternion(circle_normal, angle)

            new_obj = instanced_obj.copy()
            link_obj_to_same_collections(array_holder_obj, new_obj)
            new_obj.matrix_world = array_holder_obj.matrix_world @ Matrix.Translation(circle_pos) @ rot.to_matrix().to_4x4() @ obj_mat_rot
            new_objs.append(new_obj)

        # remove geo nodes object
        bpy.data.objects.remove(array_holder_obj)
        return new_objs

    def execute(self, context):
        found_rad_array_mod = False
        new_objs = []
        for obj in context.selected_objects:
            if obj.type == 'MESH' and get_geonodes_mod_type(obj, 'GP_RadialArray'):
                new_objs.extend(self.dissolve_radial_array_mod(context, obj))
                found_rad_array_mod = True

        if len(new_objs) > 0:
            # deselect all and select new objs
            for o in context.selected_objects:
                o.select_set(False)
            for o in new_objs:
                o.select_set(True)
            context.view_layer.objects.active = new_objs[0]

        if not found_rad_array_mod:
            self.report({'ERROR'}, "No Radial Array modifier found")
            return {'CANCELLED'}
        return {'FINISHED'}


# same thing for linear array
# use linear array offset ['Input_7'], linear array count [cnt_input], and linear array mode ['Input_8']
class GPRO_OT_DissolveLinearArray(bpy.types.Operator):
    bl_label = "Dissolve Linear Array"
    bl_idname = "group_pro.dissolve_linear_array"
    bl_description = "Dissolve Linear Array"
    bl_options = {'UNDO', 'REGISTER', 'INTERNAL'}

    @staticmethod
    def dissolve_linear_array_mod(context, linear_array_holder_obj):
        array_mod = get_geonodes_mod_type(linear_array_holder_obj, 'GP_LinearArray')
        if not array_mod:
            print("No Linear Array modifier found")
            return False
        # get inputs
        array_offset = Vector(array_mod['Input_7'])
        array_count = array_mod[cnt_input]
        array_offset_mode = array_mod['Input_8']
        instanced_obj = array_mod[inst_input]

        # decompose instanced_obj matrix_world - we will use its rotation compoenet later
        obj_mat_loc, obj_mat_rot, obj_mat_sca = matrix_decompose(instanced_obj.matrix_world)

        new_objs = []

        for i in range(array_count):
            if array_offset_mode:
                offset = array_offset * i
            else:
                offset = array_offset * i/(array_count-1)

            new_obj = instanced_obj.copy()
            link_obj_to_same_collections(linear_array_holder_obj, new_obj)
            new_obj.matrix_world = linear_array_holder_obj.matrix_world @ Matrix.Translation(offset)
            new_objs.append(new_obj)

        # remove geo nodes object
        bpy.data.objects.remove(linear_array_holder_obj)
        return new_objs

    def execute(self, context):
        found_linear_array_mod = False
        new_objs = []
        for obj in context.selected_objects:
            if obj.type == 'MESH' and get_geonodes_mod_type(obj, 'GP_LinearArray'):
                new_objs.extend(self.dissolve_linear_array_mod(context, obj))
                found_linear_array_mod = True

        if len(new_objs) > 0:
            # deselect all and select new objs
            for o in context.selected_objects:
                o.select_set(False)
            for o in new_objs:
                o.select_set(True)
            context.view_layer.objects.active = new_objs[0]

        if not found_linear_array_mod:
            self.report({'ERROR'}, "No Linear Array modifier found")
            return {'CANCELLED'}
        return {'FINISHED'}


