# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
import bpy
from .helpers_group import *


FIRST_RUN = True
#TODO: fix on opearator redo, with linked libs

class GP_OT_CreateProxy(bpy.types.Operator):
    bl_idname = "object.gp_proxy_create"
    bl_label = "Create/Regenerate Proxy"
    bl_description = "Create lowpoly version of objects inside group, and use it for drawing group in 3d view. The original objects will be hidden."
    bl_options = {"REGISTER", "UNDO"}

    reduction_method: bpy.props.EnumProperty(name='Reduction method', description='Smooth Axis',
                                             items=[('CONVEX', 'Convex Hull(Fastest)', 'Convex Hull - better Bounding Box'),
                                                    ('DECIMATE', 'Only Decimate', 'Use only Decimate modifier'),
                                                    ('REMESH', 'Remesh And Decim(Slow)', 'Use Remesh and decimate')
                                                    ], default='CONVEX')
    remesh_mode: bpy.props.EnumProperty(name="Remesh Smooth", description="", default="SMOOTH",
                                        items=(("SMOOTH", "Smooth", ""),
                                               ("SHARP", "Sharp", "")))
    octree_depth: bpy.props.IntProperty(name='octree_depth', default=6, soft_max=10, min=2)
    remove_disconnected: bpy.props.FloatProperty(name='Remove Disconnected threshold', default=0.1, min=0, max=1)

    decimate: bpy.props.FloatProperty(name='Decimate', default=0.8, min=0.01, max=1)
    # clear_mat: bpy.props.BoolProperty(name='Clear materials', default=True)


    @classmethod
    def poll(cls, context):
        ret = context.active_object and context.active_object.type == 'EMPTY' and context.active_object.instance_collection is not None or not FIRST_RUN
        return ret

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'reduction_method')
        if self.reduction_method == 'REMESH':
            layout.prop(self, 'remesh_mode')
            layout.prop(self, 'octree_depth')
            layout.prop(self, 'remove_disconnected')
        layout.prop(self, 'decimate')

    def invoke(self, context, event):
        addon_prefs = get_addon_preferences()
        self.reduction_method = addon_prefs.reduction_method
        self.remesh_mode = addon_prefs.remesh_mode
        self.octree_depth = addon_prefs.octree_depth
        self.remove_disconnected = addon_prefs.remove_disconnected
        self.decimate = addon_prefs.decimate

        if context.active_object.library or context.active_object.instance_collection.library:
            self.reduction_method = 'DECIMATE'
            self.decimate = 0.02

        return self.execute(context)

    def execute(self, context):
        global FIRST_RUN
        FIRST_RUN = False
        empty = bpy.context.active_object
        selected_coll = empty.instance_collection

        #* remove old proxy objs if exit
        rem_objs = None
        if selected_coll.gp_props.is_proxy:  # remove old objs from proxy and regenerate it
            empty.instance_collection = selected_coll.gp_props.to_parent #so that operators will work on active obj
            rem_objs = selected_coll.objects #remove objs from proxy  -- will regenerate later
            selected_coll = selected_coll.gp_props.to_parent #switch to non proxy obj
        elif hasattr(selected_coll.gp_props, 'to_proxy') and selected_coll.gp_props.to_proxy:  # sell_coll must be original. If proxy exit remove those objs
            rem_objs = selected_coll.gp_props.to_proxy.objects
        if rem_objs:
            for obj in reversed(rem_objs):
                bpy.data.objects.remove(obj)

        #* dupilcate group, make unique, convert_to geo, applpy mods, remove non mesh objs
        bpy.ops.object.duplicate()

        #!belower crashes on redo on empty library obj
        # obj_before = [o for o in context.visible_objects]
        # bpy.ops.object.duplicates_make_real()
        # bpy.ops.object.make_local(type='SELECT_OBDATA')

        # for o in context.visible_objects:
        #     if o not in obj_before:
        #         o.select_set(True)


        bpy.ops.object.gpro_makeunique(maxDepth=0, makeDataUnique=1, makeMaterialsUnique=0)
        bpy.ops.object.gpro_dissolve(maxDepth=0, dissolve_modifiers=True) # XXX: geonodes modifiers may be dissolved into instance objs - these may be still linked

        # new
        bpy.ops.object.make_single_user(object=True, obdata=True, material=False, animation=False, obdata_animation=False)

        #* make active one mesh obj, from dissolved group
        for obj in context.selected_objects:
            if obj.type == 'MESH':
                context.view_layer.objects.active = obj
                proxy_obj_low = obj
                break
        # bpy.ops.object.make_single_user(type='SELECTED_OBJECTS')

        bpy.ops.object.convert(target='MESH') #Apply mods
        #* delete bbox obj, or non mesh obj (we do not need them in proxy)
        del_objs = [obj for obj in context.selected_objects if obj.type != 'MESH' or not obj.visible_get() or obj.display_type == 'BOUNDS']
        for obj in reversed(del_objs):
            bpy.data.objects.remove(obj)

        if len(context.selected_objects)>1:
            bpy.ops.object.join()
        tris_before = len(proxy_obj_low.data.polygons)

        #* Decimate proxy objs
        if self.reduction_method == 'REMESH':
            bpy.ops.object.modifier_add(type="REMESH")
            bpy.context.object.modifiers["Remesh"].mode = self.remesh_mode
            bpy.context.object.modifiers["Remesh"].octree_depth = self.octree_depth
            if self.remove_disconnected:
                bpy.context.object.modifiers["Remesh"].threshold = self.remove_disconnected
            else:
                bpy.context.object.modifiers["Remesh"].use_remove_disconnected = False

        elif self.reduction_method == 'CONVEX':
            bpy.ops.object.editmode_toggle()
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.convex_hull()
            bpy.ops.object.editmode_toggle()


        bpy.ops.object.modifier_add(type="DECIMATE")
        bpy.context.object.modifiers["Decimate"].ratio = 1-self.decimate
        bpy.ops.object.convert(target='MESH')

        #* setup proxy collection, unlink proxy objs from scene, etc
        proxy_obj_low.matrix_world = empty.matrix_world.inverted() @ proxy_obj_low.matrix_world
        proxy_coll = selected_coll.gp_props.to_proxy if selected_coll.gp_props.to_proxy else bpy.data.collections.new(selected_coll.name + '_proxy')
        proxy_coll.gp_props.is_proxy = True

        proxy_coll.objects.link(proxy_obj_low)
        proxy_coll.instance_offset = selected_coll.instance_offset

        selected_coll.id_data.user_remap(proxy_coll) #replace sel_coll with proxy
        proxy_coll.gp_props.to_parent = selected_coll  #! wtf this crashes
        selected_coll.gp_props.to_proxy = proxy_coll

        proxy_coll.created_with_gp = True
        selected_coll.use_fake_user = True
        getGroupBBoxVerts(empty)
        # selected_coll.use_fake_user = True
        # link_obj_to_same_collections(empty, current_group)
        # if self.clear_mat:
        #     proxy_obj_low.data.materials.clear()
        unlink_obj_from_scene(context, proxy_obj_low, exclude_coll=proxy_coll)
        empty.select_set(True)
        context.view_layer.objects.active = empty
        self.report({'INFO'}, f'Tris before: {tris_before}, now: {len(proxy_obj_low.data.polygons)}')
        return {"FINISHED"}


class GP_OT_ProxySwitch(bpy.types.Operator):
    bl_idname = "object.gp_proxy_switch"
    bl_label = "Proxy switch"
    bl_description = "Switch to proxy or to origina object"
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'EMPTY' and context.active_object.instance_collection

    def execute(self, context):
        empty = context.active_object
        # current_empty_m_w.translation = Vector((0,0,0))
        selected_coll = empty.instance_collection

        switch_coll = None
        if selected_coll.gp_props.to_proxy:
            # selected_coll.use_fake_user = True  # use when removing direct link to
            switch_coll = selected_coll.gp_props.to_proxy
        elif selected_coll.gp_props.is_proxy:
            switch_coll = selected_coll.gp_props.to_parent
            # switch_coll.use_fake_user = False #no need when loading to scene.

        if not switch_coll:
            if selected_coll.gp_props.is_proxy:
                self.report({'ERROR'}, 'Original group link lost! Cancelling')
            else:
                self.report({'ERROR'}, 'There is not proxy for selected group! Cancelling')
            return {'CANCELLED'}

        # selected_coll.id_data.user_remap(switch_coll)  #! remaps orig_col.to proxy to self => cyclic dep!
        for obj in bpy.data.objects:
            if obj.type == 'EMPTY' and obj.instance_collection == selected_coll:
                obj.instance_collection = switch_coll
        context.area.tag_redraw()
        return {"FINISHED"}


class GP_OT_DeleteProxy(bpy.types.Operator):
    bl_idname = "object.gp_delete_proxy"
    bl_label = "Delete proxy"
    bl_description = "Delete proxy version of group"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'EMPTY' and context.active_object.instance_collection

    @staticmethod
    def remove_proxy(empty):
        # current_empty_m_w.translation = Vector((0,0,0))
        selected_coll = empty.instance_collection

        if selected_coll.gp_props.is_proxy:
            proxy_coll = selected_coll
            selected_coll = selected_coll.gp_props.to_parent
            if not selected_coll:
                return ['ERROR','Original group link lost! Cancelling']
            bpy.ops.object.gp_proxy_switch()
        else:
            selected_coll = empty.instance_collection
            proxy_coll = selected_coll.gp_props.to_proxy

        if proxy_coll:
            for obj in reversed(proxy_coll.objects):
                bpy.data.objects.remove(obj)
            proxy_name = proxy_coll.name
            bpy.data.collections.remove(proxy_coll)
            return ['INFO', f'Proxy: {proxy_name} was removed']
        else:
            return ['WARNING', 'There is not proxy! Cancelling']


    def execute(self, context):
        empty = context.active_object
        info = self.remove_proxy(empty)
        if info:
            self.report({info[0]}, info[1])
        return {"FINISHED"}
