# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
import os
import bmesh
from pathlib import Path
import numpy as np
from .tree_structure import TreeData
from mathutils import Vector, Matrix, Color


def is_inst_coll(obj):
    return obj and obj.type == 'EMPTY' and obj.instance_collection

#########    Global functions      ###################################

def ConvertMatrixToList(mat):
    return tuple([b for a in mat for b in a])


def ListToMat4D(inputList):
    out = ()
    for i in range(4):
        out = out + (tuple(inputList[i * 4:(i * 4 + 4)]),)
    return tuple(out)


def getRidOfLastDigits(text):
    if len(text) > 3:
        if text[-4] == '.':
            if text[-3:].isdigit():
                text = text[:-4]
    if text[0:3] == 'GI_':
        text = text[3:]
    return text


def getGroupSizeForEmpty(empty):
    if empty.type == 'EMPTY' and empty.instance_type == 'COLLECTION':
        center = empty.instance_collection.instance_offset * (-1)
        total_minx = 0.0
        total_maxx = 0.0
        total_miny = 0.0
        total_maxy = 0.0
        total_minz = 0.0
        total_maxz = 0.0
        for obj in empty.instance_collection.objects:
            scale = obj.scale
            minx = obj.bound_box[0][0] * scale.x + obj.location[0]
            total_minx = min(minx, total_minx)
            maxx = obj.bound_box[4][0] * scale.x + obj.location[0]
            total_maxx = max(maxx, total_maxx)
            miny = obj.bound_box[0][1] * scale.y + obj.location[1]
            total_miny = min(miny, total_miny)
            maxy = obj.bound_box[2][1] * scale.y + obj.location[1]
            total_maxy = max(maxy, total_maxy)
            minz = obj.bound_box[0][2] * scale.z + obj.location[2]
            total_minz = min(minz, total_minz)
            maxz = obj.bound_box[1][2] * scale.z + obj.location[2]
            total_maxz = max(maxz, total_maxz)

        xDist = max(abs(total_maxx - center[0]), abs(total_minx - center[0]))
        yDist = max(abs(total_maxy - center[1]), abs(total_miny - center[1]))
        zDist = max(abs(total_maxz - center[2]), abs(total_minz - center[2]))

        return max(xDist, yDist, zDist)
    return 1


def get_coll_bbox_fast(coll, depsgraph, ignore_visible = True):
    ''' usually seems way faster than get_nested_obj_depsgrapth_bbox..., unless it calls get_nested_obj_depsgrapth_bbox ? I think '''
    min_x, max_y = 10000.0, -10000.0
    min_y, max_x = 10000.0, -10000.0
    min_z, max_z = 10000.0, -10000.0

    slow_deps_bbox = []
    for obj in coll.all_objects:
        if obj.type == 'EMPTY' and obj.instance_type == 'COLLECTION' and (ignore_visible or obj.visible_get()):  # if group has empty with instance_collection ten get bbxo from instance_collection BBoxVertsList
            bBoxCoordList = list(obj.instance_collection.BBoxVertsList)
            for i in range(0, 8):  # for box vert list  3*8   (x,y,z)*8
                x,y,z = obj.matrix_world@Vector((bBoxCoordList[i*3], bBoxCoordList[i*3+1], bBoxCoordList[i*3+2]))  # decode list to vers
                if x < min_x:
                    min_x = x
                if x > max_x:
                    max_x = x
                if y < min_y:
                    min_y = y
                if y > max_y:
                    max_y = y
                if z < min_z:
                    min_z = z
                if z > max_z:
                    max_z = z

        elif (ignore_visible or obj.visible_get()):
            empty_bbox = (Vector(obj.bound_box[6])-Vector(obj.bound_box[0])).length # corners across the diagonal
            if empty_bbox > 0.0001:
                for vert in obj.bound_box:
                    x, y, z = obj.matrix_world@Vector(vert) # if bbox are all zeroes - then this will basically output obj center position
                    if x < min_x:
                        min_x = x
                    if x > max_x:
                        max_x = x
                    if y < min_y:
                        min_y = y
                    if y > max_y:
                        max_y = y
                    if z < min_z:
                        min_z = z
                    if z > max_z:
                        max_z = z
            else: # eg when obj has no verts, -- eg. on gpro rad array, or linear array
                slow_deps_bbox.append(obj)

    if slow_deps_bbox: # thx to this, iter over all depsgraph objects - only once
        # WARN: may be slow since it iterates over all generated eval.objects - for each obj in coll loop... quadratic complexity..
        (lmin_x, lmax_x), (lmin_y, lmax_y), (lmin_z, lmax_z) = get_nested_obj_depsgrapth_bbox(depsgraph, slow_deps_bbox, apply_modifiers=True) # slow - last resort
        if lmin_x < min_x:
            min_x = lmin_x
        if lmax_x > max_x:
            max_x = lmax_x
        if lmin_y < min_y:
            min_y = lmin_y
        if lmax_y > max_y:
            max_y = lmax_y
        if lmin_z < min_z:
            min_z = lmin_z
        if lmax_z > max_z:
            max_z = lmax_z

    return (min_x, max_x), (min_y, max_y), (min_z, max_z)



def getEditedBBoxVerts(context):  # calculates edited group bb vertices
    storageList = context.scene.storedGroupSettings
    edited_empty = bpy.data.objects[storageList[-1].currentEmptyName]

    context = bpy.context
    coll = edited_empty.instance_collection
    if not coll:
        return None

    depsgraph = context.evaluated_depsgraph_get()
    #  bit to slow for realtime drawing - since it loos on all depsgraph objects - their meshes..
    # (min_x, max_x), (min_y, max_y), (min_z, max_z) = get_nested_obj_depsgrapth_bbox(depsgraph, coll.objects[:], apply_modifiers=True)


    # NOTE: restored old, simpler way -- is it?
    (min_x, max_x), (min_y, max_y), (min_z, max_z) = get_coll_bbox_fast(coll, depsgraph, ignore_visible = False) # seem faster


    edges = [ [0, 1], [1, 2], [2, 3], [3, 0], [4, 5], [5, 6], [6, 7], [7, 4], [0, 4], [1, 5], [2, 6], [3, 7] ]

    # create bb verts top face 4 then bottom 4
    verts = [(max_x, min_y, max_z),
                (max_x, max_y, max_z),
                (min_x, max_y, max_z),
                (min_x, min_y, max_z),
                (max_x, min_y, min_z),
                (max_x, max_y, min_z),
                (min_x, max_y, min_z),
                (min_x, min_y, min_z)
                ]

    return [verts[vert] for edge in edges for vert in edge]


def calc_objs_bb_bottom_center(objs):
    context = bpy.context
    depsgraph = context.evaluated_depsgraph_get()

    (min_x, max_x), (min_y, max_y), (min_z, max_z) = get_nested_obj_depsgrapth_bbox(depsgraph, objs, apply_modifiers=True)
    return Vector(((max_x + min_x) / 2, (max_y + min_y) / 2, min_z))


def get_nested_obj_depsgrapth_bbox(depsgraph, src_objs, apply_modifiers=True, matrix=Matrix.Identity(4)):
    ''' This is precise way to get bbox but slow, - if we loop for bunch of src_obj,  not for view drawing - supper laggy'''
    # TODO: if not selected - eg when closing all groups, and edited group is in other scene, this seems to give 0,0,0 - fix
    # TODO: way faster method is to use src_obj's - and loop once over  depsgraph.object_instances. Then check if object_instance.parent.original in src_obj's
    def is_object_instance_from_selected(object_instance, base_objs):
        # For instanced objects we check selection of their instancer (more accurately: check
        # selection status of the original object corresponding to the instancer).
        if object_instance.parent: #  it refers to group instance, rather than  rad array modifier.
            return object_instance.parent.original in base_objs
        # For non-instanced objects we check selection state of the original object.
        return object_instance.object.original in base_objs

    def get_obj_for_instance(object_instance, apply_modifiers):
        if apply_modifiers:
            return object_instance.object
        else:
            return object_instance.object.original

    '''Works in groups too in recursive way. apply_modifiers=False => modes get ignored competly'''
    min_x, max_y = 10000.0, -10000.0
    min_y, max_x = 10000.0, -10000.0
    min_z, max_z = 10000.0, -10000.0
    m_inv = matrix.inverted()
    for dep_obj_info in depsgraph.object_instances:
        if not is_object_instance_from_selected(dep_obj_info, src_objs):
            continue # We only export selected objects

        orig_obj = get_obj_for_instance(dep_obj_info, apply_modifiers)
        mat = m_inv @ dep_obj_info.matrix_world
        if not orig_obj.data or orig_obj.type not in ('MESH','CURVE','CURVES'): # for non-geometry objects, use their location center
            x, y, z = mat @ Vector((0, 0, 0)) # translation already included in matrix
            if x < min_x:
                min_x = x
            if x > max_x:
                max_x = x
            if y < min_y:
                min_y = y
            if y > max_y:
                max_y = y
            if z < min_z:
                min_z = z
            if z > max_z:
                max_z = z

        else:
            for vert in orig_obj.bound_box:
                x, y, z = mat @ Vector(vert)
                if x < min_x:
                    min_x = x
                if x > max_x:
                    max_x = x
                if y < min_y:
                    min_y = y
                if y > max_y:
                    max_y = y
                if z < min_z:
                    min_z = z
                if z > max_z:
                    max_z = z

    return (min_x, max_x), (min_y, max_y), (min_z, max_z)

def get_obj_bbox_manually(obj):
    '''Works in groups too in recursive way. apply_modifiers=False => modes get ignored competly'''
    if len(obj.data.vertices) == 0:
        return Vector((0,0,0))
    tempx = []
    tempy = []
    tempz = []
    mw = obj.matrix_world
    for vert in obj.data.vertices:
        x, y, z = mw @ Vector(vert.co)
        tempx.append(x)
        tempy.append(y)
        tempz.append(z)
    x_dim = max(tempx) - min(tempx)
    y_dim = max(tempy) - min(tempy)
    z_dim = max(tempz) - min(tempz)
    return Vector((x_dim, y_dim, z_dim))


def getGroupBBoxVerts(empty):
    coll = empty.instance_collection
    context = bpy.context
    depsgraph = context.evaluated_depsgraph_get()

    (min_x, max_x), (min_y, max_y), (min_z, max_z) = get_nested_obj_depsgrapth_bbox(depsgraph, [empty], apply_modifiers=True, matrix=empty.matrix_world) # ok, but ignores eg. curves or metaballs

    coll.BBoxVertsList = [max_x, min_y, max_z,
           max_x, max_y, max_z,
           min_x, max_y, max_z,
           min_x, min_y, max_z,
           max_x, min_y, min_z,
           max_x, max_y, min_z,
           min_x, max_y, min_z,
           min_x, min_y, min_z]

    if get_addon_preferences().disable_bbox_verts:
        update_all_bb_objs() #clears bbmesh (and objs) of disabled
    else:
        sizeVec, center = get_bbox_size_and_center(coll)
        setup_bb_object(coll, center, sizeVec)

def get_bbox_size_and_center(coll):
    verts = coll.BBoxVertsList
    sizeVec = Vector((verts[0]-verts[6], verts[4]-verts[1], verts[2]-verts[-1]))
    center = Vector(((verts[0]+verts[6])/2, (verts[4]+verts[1])/2, (verts[2]+verts[-1])/2))
    return sizeVec, center

def remove_broken_edit_groups(self, context):
    for idx, storage in reversed(list(enumerate(context.scene.storedGroupSettings))):
        currentEmpty = bpy.data.objects.get(storage.currentEmptyName)
        if not currentEmpty:
            self.report({'WARNING'}, message=f"Edited group: {idx} is broken. Removing it from edit stack")
            context.scene.storedGroupSettings.remove(idx)
            continue
        current_group = currentEmpty.instance_collection
        if not current_group:
            self.report({'WARNING'}, message=f"Edited group: {idx} is broken. Removing it from edit stack")
            context.scene.storedGroupSettings.remove(idx)

inv_g = 1.0/1.6180339887498948482 # 1/gold_ratio
def gold_random(i, seed):
    ''' seed in (0,1) range '''
    return (seed + i*inv_g) % 1

def update_all_coll_colors():
    prefs = get_addon_preferences()
    if prefs.auto_color:
        # it would be better to use blend file property to store was color overridden  but no option in blender
        if bpy.context.scene.gp_props.viewcolor_override_enabled == False: # did we override already in past? then skip
            for coll in bpy.data.collections:
                set_coll_viewport_color(coll, reset=False)
        bpy.context.scene.gp_props.viewcolor_override_enabled = True
    else:
        if bpy.context.scene.gp_props.viewcolor_override_enabled == True:  # reset colors if were overriden
            for coll in bpy.data.collections:
                set_coll_viewport_color(coll, reset=True)
        bpy.context.scene.gp_props.viewcolor_override_enabled = False


def set_coll_viewport_color(coll, reset=False):
    if not reset:
        i = abs(hash(coll.id_data))%999999
        # np.random.seed(seed)
        # gold_seed = np.random.rand()  # (0,1) range
        rand_color = Color((0.7, 0.7, 0.7))  # h, sat, valu
        rand_color.s = 0.5
        rand_color.h = gold_random(i, 0)
    for obj in coll.users_dupli_group:
        if obj.type == 'EMPTY':
            if reset:
                obj.color = obj.gp_props.orig_viewport_color
            else:
                obj.gp_props.orig_viewport_color = obj.color
                obj.color[0:3] = rand_color


def setup_bb_object(instance_collection, bb_center, bb_scale):
    current_coll = instance_collection
    if current_coll.gp_props.bbox_obj:
        gp_bbox_object = current_coll.gp_props.bbox_obj
    else:
        mesh = get_set_bb_mesh()
        gp_bbox_object = bpy.data.objects.new('gp_bbox_object', mesh)
        gp_bbox_object.display.show_shadows = False
        gp_bbox_object.display_type = 'WIRE'  # avoid drawign ugly dots
        gp_bbox_object.hide_render = True
        gp_bbox_object.use_fake_user = True
        current_coll.gp_props.bbox_obj = gp_bbox_object

    if gp_bbox_object.name not in current_coll.objects.keys():  # cos Edit group, unlinks it from edited_coll
        current_coll.objects.link(gp_bbox_object)
    mat_loc = Matrix.Translation(bb_center)
    mat_scale = get_scale_mat(bb_scale/2)
    gp_bbox_object.matrix_basis = mat_loc @ mat_scale
    print(f'Added bbox to {instance_collection.name} collection')
    # gp_bbox_object.location = bb_center
    # gp_bbox_object.scale = bb_scale
    # gp_bbox_object.rotation_euler = bb_scale


def update_all_bb_objs():
    bb_mesh_exist = bpy.data.meshes.get('gp_bbox_mesh') #if exist -then objects must use Gpro BBox mesh
    disable_bbox_verts = get_addon_preferences().disable_bbox_verts
    if disable_bbox_verts:
        any_bb_enabled = any([scn.gp_props.bbox_enabled for scn in bpy.data.scenes])
        if bb_mesh_exist:
            for coll in bpy.data.collections:
                bbox_obj = coll.gp_props.bbox_obj
                if bbox_obj and bbox_obj.data:
                    bpy.data.meshes.remove(bbox_obj.data) #! this also removes bb objs
                    print(f'Removed bbox object from {coll.name} collection')

            for scn in bpy.data.scenes:
                scn.gp_props.bbox_enabled = False

            # bpy.data.meshes.remove(bb_mesh_exist)
            print(f"Removed GropPro Bounding Box mesh (see 'Disable BBox Vertsices' option in GroupPro preferences)")

    else: #* enable bbox to all collections
        checked_collections = []
        def set_bbox_rec(current_group):
            if current_group.name in checked_collections:  # skip if already checked
                return
            else:
                if not current_group.gp_props.bbox_obj:
                    sizeVec, center = get_bbox_size_and_center(current_group)
                    setup_bb_object(current_group, center, sizeVec)
                checked_collections.append(current_group.name)

            for obj in current_group.objects:
                if obj.instance_collection and obj.type == 'EMPTY':
                    set_bbox_rec(obj.instance_collection)

            for child_coll in current_group.children:#* add sub colls support
                for obj in child_coll.all_objects:
                    if obj.type == 'EMPTY' and obj.instance_collection:
                        set_bbox_rec(obj.instance_collection)
                checked_collections.append(child_coll.name)

        for scn in bpy.data.scenes:
            if not scn.gp_props.bbox_enabled:
                scn.gp_props.bbox_enabled = True
                for obj in scn.objects:
                    if obj.instance_collection and obj.type == 'EMPTY':
                        set_bbox_rec(obj.instance_collection)


def get_set_bb_mesh():
    # verts = [Vector((1, -1, 1)),
        #         Vector((1, 1, 1)),
        #         Vector((-1, 1, 1)),
        #         Vector((-1, -1, 1)),
        #         Vector((1, -1, -1)),
        #         Vector((1, 1, -1)),
        #         Vector((-1, 1, -1)),
        #         Vector((-1, -1, -1))]
        # edges = []
        # #as quads
        # # bbox_faces = [(1, 0, 4, 5), (4, 0, 3, 7), (2, 3, 0, 1), (6, 2, 1, 5), (6, 5, 4, 7), (2, 6, 7, 3)]
        # #as tris
        # # bbox_faces = [(4, 6, 5), (0, 1, 2), (1, 6, 2), (4, 1, 0), (2, 7, 3), (3, 4, 0), (4, 7, 6), (0, 2, 3), (1, 5, 6), (4, 5, 1), (2, 6, 7), (3, 7, 4)]
        # bbox_faces = []
        # mesh = bpy.data.meshes.new(name="gp_bbox_mesh")
        # mesh.use_fake_user = True
        # mesh.from_pydata(verts, edges, bbox_faces)
    bb_mesh = bpy.data.meshes.get('gp_bbox_mesh')
    if bb_mesh:
        if len(bb_mesh.vertices) == 8:
            return bb_mesh
        else:      # must be broken. Recreate it

            bb_mesh.clear_geometry()
    else:
        bb_mesh = bpy.data.meshes.new(name="gp_bbox_mesh")
    bb_mesh.use_fake_user = True
    verts_co = [Vector((1, -1, 1)),
                Vector((1, 1, 1)),
                Vector((-1, 1, 1)),
                Vector((-1, -1, 1)),
                Vector((1, -1, -1)),
                Vector((1, 1, -1)),
                Vector((-1, 1, -1)),
                Vector((-1, -1, -1))]
    bm = bmesh.new()   # create an empty BMesh
    for v_co in verts_co:
        bm.verts.new(v_co)
    bm.to_mesh(bb_mesh)
    bm.free()
    return bb_mesh




def localView_up(self, context): #only for calling from props update function
    gp_localView(self, context)


def get_3d_view():
    areas = bpy.context.workspace.screens[0].areas
    for area in areas:
        for space in area.spaces:
            if space.type == 'VIEW_3D':
                return space

def gp_localView(self, context, reveal_in_local = []):
    gp_prefs = get_addon_preferences()
    # support for local view from outliner - create group (context has no space_data.local_view then ) ? Or just return...
    space_data = context.space_data if hasattr(context.space_data, 'local_view') else get_3d_view()

    if len(context.scene.storedGroupSettings) == 0: #nothing to isolate
        for obj in context.scene.objects:  # edit current empty objects
            if obj.gp_props.gpro_hide:  # if hidden by GroupPro
                obj.hide_viewport = False
                obj.gp_props.gpro_hide = False


        if space_data.local_view:  # reveal in local view
            for obj in reveal_in_local:
                obj.local_view_set(space_data, True)
            local_view_objs = {obj for obj in context.scene.objects if obj.local_view_get(space_data)} if space_data.local_view else set()
            if len(local_view_objs) == 0:
                bpy.ops.view3d.localview(frame_selected=False)
        return

    groupTree = TreeData()  # queue of sets of groups - just for computing depth

    # building tree datablock of edited groups with nesting
    for storage in context.scene.storedGroupSettings:
        empty = bpy.data.objects.get(storage.currentEmptyName)
        if not empty:
            continue
        currentlyEditedGroup = empty.instance_collection  # meaning currentEmpty Is Parent Of This Group
        #TODO: when empty is unlinkted it is removed from collection, for edit time. So thism makes users_collection==0 alway. Work around?
        parent_groups = [col.name for col in bpy.data.collections if currentlyEditedGroup.name in col.children.keys()]
        if len(parent_groups) > 0:
            for parent_group in parent_groups:
                groupTree.SearchAndInsert(parent_group, currentlyEditedGroup.name)
        else:
            groupTree.SearchAndInsert('', currentlyEditedGroup.name)

    # get deepest leaf in tree datablock -MaxDepth how nested we are in reality?
    group_objs_to_reveal = set()
    group_objs_to_hide = set()
    MaxDepth = groupTree.GetMaxDepth() if gp_prefs.isolate_mode == 'TREE' else len(context.scene.storedGroupSettings)  # max depth of edited nested groups
    # get each edited group 'real Depth' groupDepth
    for i,storage in enumerate(context.scene.storedGroupSettings):
        empty = bpy.data.objects.get(storage.currentEmptyName)
        if not empty:
            continue
        currentlyEditedGroup = empty.instance_collection.name  # meaning currentEmpty Is Parent Of This Group
        if gp_prefs.isolate_mode == 'TREE':
            groupDepth = groupTree.FindElementDepth(currentlyEditedGroup)  # search 'real Depth' of currently edited  group
        else:
            groupDepth = i + 1  # search 'real Depth' of currently edited  group
        # print('group '+currentlyEditedGroup+' Depth is '+str(groupDepth))
        editedGroup = empty.instance_collection
        if groupDepth > MaxDepth-context.scene.gp_props.groupLocalViewDepth:
            group_objs_hidden = [o.name for o in editedGroup.all_objects if not o.hide_viewport or o.gp_props.gpro_hide] #visible stay visible, hiden by gpro - unhide too
            group_objs_to_reveal |= set(group_objs_hidden)
        else:
            group_objs_visible = [o.name for o in editedGroup.all_objects if not o.hide_viewport] #bisible then hide
            group_objs_to_hide |= set(group_objs_visible)
# mark all obj from group above depth for hide

    groupTree.clear()

    if context.scene.gp_props.groupLocalView:
        for obj in context.scene.objects:  # edit current empty objects
            if obj.name in group_objs_to_reveal:
                obj.hide_viewport = False  # revala edited obj
                obj.gp_props.gpro_hide = False
            elif obj.name in group_objs_to_hide:
                obj.hide_viewport = True
                obj.gp_props.gpro_hide = True
            elif not obj.hide_viewport:  # hide remaining viisble objs
                obj.hide_viewport = True
                obj.gp_props.gpro_hide = True
                # skip if obj was already hidden - prevents unhiding its later

    else:  # unhide object hidded by GP
        for obj in context.scene.objects:
            if obj.gp_props.gpro_hide:
                obj.hide_viewport = False
                obj.gp_props.gpro_hide = False

    if space_data.local_view:  # reveal in local view
        for obj in reveal_in_local:
            obj.local_view_set(space_data, True)
        local_view_objs = {obj for obj in context.scene.objects if obj.local_view_get(space_data)} if space_data.local_view else set()
        if len(local_view_objs) == 0:
            bpy.ops.view3d.localview(frame_selected=False)

    # remove stored group settings if  group empty was deleted
    for storage in reversed(context.scene.storedGroupSettings):
        if not bpy.data.objects.get(storage.currentEmptyName):
            context.scene.storedGroupSettings.remove(len(context.scene.storedGroupSettings)-1)
    group_objs_to_reveal.clear()


def operator_is_redo(cls, context):
    '''True if we executed operator, and we are still in operator redo panel'''
    op = context.active_operator
    return op and (getattr(op.bl_rna,'bl_idname',None) == cls.bl_idname)

def check_file_exist(filePath):
    """Retuns absolute file path, and bool - file exist?"""
    abspathToFix = Path(bpy.path.abspath(filePath))  # crappy format like c:\\..\\...\\ddada.fbx
    outputPathStr = str(abspathToFix.resolve())
    if abspathToFix.is_dir():
        outputPathStr += '\\'
    return outputPathStr, os.path.isfile(outputPathStr)

def get_exchange_filepath():
    addon_prefs = get_addon_preferences()
    if addon_prefs.exchange_path_type == "BLEND":
        out_path = bpy.path.abspath("//")  # bpy.path.abspath("//LibExchange\\")
    elif addon_prefs.exchange_path_type == "CUSTOM":  # exchange path + blend file name
        out_path = bpy.path.abspath(addon_prefs.exchange_path + "\\")
    if not os.path.exists(out_path):
        os.makedirs(out_path)
    return out_path + 'gp_lib_data.blend'


def get_edited_colls(lib = None):
    storageList = bpy.context.scene.storedGroupSettings
    if lib:
        edited_col = []
        for storage in storageList:
            col = bpy.data.objects[storage.currentEmptyName].instance_collection
            if col:
                if col.library == lib or ('lib_path' in col and col['lib_path'] == lib.filepath):
                    edited_col.append(col)
    else:
        edited_col = [bpy.data.objects[storage.currentEmptyName].instance_collection for storage in storageList if bpy.data.objects.get(storage.currentEmptyName)]
    return edited_col


def BC_check():
    return 'BoxCutter' in bpy.context.preferences.addons.keys()

def get_addon_name():
    return __package__

def addon_name_lowercase():
    return get_addon_name().lower()

def get_addon_preferences():
    return bpy.context.preferences.addons[get_addon_name()].preferences

def clear_parent(obj):
    parent = obj.parent
    if parent:
        if obj.parent_type == "OBJECT":
            obj.matrix_basis = parent.matrix_world @ obj.matrix_parent_inverse @ obj.matrix_basis
            obj.parent = None
        elif obj.parent_type == 'BONE':
            trans_mat = Matrix.Identity(4)
            if obj.parent_bone:
                trans_mat = obj.parent.pose.bones[obj.parent_bone].matrix
            trans_mat = parent.matrix_world @ trans_mat
            obj.matrix_basis = trans_mat @ obj.matrix_parent_inverse @ obj.matrix_basis
            obj.parent = None

def set_parent(obj, new_parent):
    if obj.parent:
        old_parent_m_w = obj.parent.matrix_world.copy() #backup cos we will change obj.parent below, so backup
        backup_obj_mpi = obj.matrix_parent_inverse.copy()  # bug in blender: obj.parent = new_parent clears obj.matrix_parent_inverse so backup it
        obj.parent = new_parent
        new_m_p_inv = new_parent.matrix_world.inverted() @ old_parent_m_w @ backup_obj_mpi
        obj.matrix_parent_inverse = new_m_p_inv
    else:
        obj.parent = new_parent
        obj.matrix_parent_inverse = new_parent.matrix_world.inverted()

def deleteGroup(currentGroup):
    for groupObj in currentGroup.objects:
        bpy.data.objects.remove(groupObj)
    for obj in bpy.data.objects:    # then delete all empties in scene that used that group  -- including currentEmpty (which is now linked to scene)
        # or if different empties are left empty, then delete those too
        if obj.type == 'EMPTY' and obj.instance_collection and (obj.instance_collection.name == currentGroup.name or len(obj.instance_collection.objects) == 0):
            bpy.data.objects.remove(obj)
    bpy.data.collections.remove(currentGroup, do_unlink=True)

def link_obj_to_same_collections(source_obj, clone):
    ''' Link obj to collections where source_obj is located '''
    for source_obj_coll_parent in source_obj.users_collection:
        if clone.name not in source_obj_coll_parent.objects.keys():
            source_obj_coll_parent.objects.link(clone)

    if source_obj.name in bpy.context.scene.collection.objects.keys():
        if clone.name not in bpy.context.scene.collection.objects.keys():
            bpy.context.scene.collection.objects.link(clone)

def link_coll_to_same_collections(source_obj, clone_coll):
    ''' Link clone col to collections where source_obj is located '''
    for source_obj_coll_parent in source_obj.users_collection:
        if clone_coll.name not in source_obj_coll_parent.children.keys():
            source_obj_coll_parent.children.link(clone_coll)
    if source_obj.name in bpy.context.scene.collection.objects.keys():
        if clone_coll.name not in bpy.context.scene.collection.children.keys():
            bpy.context.scene.collection.children.link(clone_coll)

def find_create_collection(parent_col, searched_col_name):
    ''' Try to find searched collection name in parent childrens. If not found create it and link it to master'''
    def find_coll(p_coll, search_col):
        if search_col == p_coll.name:
            return True
        if search_col in p_coll.children.keys():
            return True
        else:
            for child_col in p_coll.children:
                if find_coll(child_col, search_col):
                    return True
        return False

    if find_coll(parent_col, searched_col_name):
        collection = bpy.data.collections[searched_col_name]
    else:
        if searched_col_name in bpy.data.collections.keys():
            collection = bpy.data.collections[searched_col_name]
        else:
            collection = bpy.data.collections.new(searched_col_name)
        parent_col.children.link(collection)
    return collection

def get_view_layer_collection_tree(context, searched_obj):
    ''' return list of collections that array_obj is in'''
    collection_tree = []
    def get_collection_tree(current_col): # use view_layer collections
        if current_col.collection.objects.get(searched_obj.name):
            collection_tree.append(current_col)
            return True

        for sub_col in current_col.children:
            if get_collection_tree(sub_col): # if array obj is in sub_col
                collection_tree.append(current_col) # add current col to tree
                return True

        return False
    get_collection_tree(context.view_layer.layer_collection) # start from master/scene collection
    return collection_tree


def unlink_obj_from_scene(context, obj, exclude_coll = None):
    ''' ulink form all collections, exept exclude_coll'''
    # obj.use_fake_user = True
    for coll in obj.users_collection:
        if exclude_coll and coll == exclude_coll:
            continue
        coll.objects.unlink(obj)  # needs to be visible for matrix transform to work so we hide it now
    if obj.name in context.scene.collection.objects.keys():  # if in master collection unlinke it
        context.scene.collection.objects.unlink(obj)

def link_to_scene_coll(context, obj):
    if obj.name not in context.scene.collection.objects.keys():
        context.scene.collection.objects.link(obj)


def find_layer_coll(layerColl, coll):
    found = None
    if (layerColl.collection == coll):
        return layerColl
    for layer in layerColl.children:
        found = find_layer_coll(layer, coll)
        if found:
            return found
    return None

def set_active_layer_coll(context, target_coll):
    found_layer_coll = find_layer_coll(context.view_layer.layer_collection, target_coll)
    if found_layer_coll:
        context.view_layer.active_layer_collection = found_layer_coll


def unlink_col_from_coll_children_nested(parent_col, remove_coll):
    '''unlink remove_coll from parent_col.children recursive '''
    child_objs = []
    for child_coll in parent_col.children:  # and its child cools
        unlink_col_from_coll_children_nested(child_coll, remove_coll)
    if remove_coll.name in parent_col.children.keys():
        parent_col.children.unlink(remove_coll)


def coll_instances_empties(current_group):
    ''' yield all instance coll empties, with coll == current_group'''
    for obj in bpy.data.objects:
        if obj.type == 'EMPTY' and obj.instance_collection and obj.instance_collection.name == current_group.name:
            yield obj

def apply_mat_to_inner_coll_objs(current_group, mat):
    '''analogous to  apply obj.data.transform - but to collection inner objs'''
    quaternion_rot = get_addon_preferences().quaternion_rot
    active_coll_nested_objs = current_group.all_objects
    children = {obj for inner_obj in active_coll_nested_objs for obj in inner_obj.children} #set
    childrenInsidesideGroup = children.intersection(set(active_coll_nested_objs))  # selected objs, with childrent inside group

    # negate matrix transformatin on all other empties instances (including empty).
    for group_member_obj in active_coll_nested_objs:
        if group_member_obj not in childrenInsidesideGroup and group_member_obj != current_group.gp_props.bbox_obj:  # do not move children, since they will inherit mat
            if quaternion_rot:
                group_member_obj.rotation_mode = 'QUATERNION'
            else:
                group_member_obj.rotation_mode = 'XYZ'
            group_member_obj.matrix_basis = mat @ group_member_obj.matrix_basis

            if group_member_obj.gp_props.flip_props.origin == 'CURSOR': #fix 'flip' operator mirror point if using cursor
                fix_mat = mat @ group_member_obj.gp_props.flip_props.cursor_mat
                group_member_obj.gp_props.flip_props.cursor_mat = np.array(fix_mat).transpose().ravel() #flatten

    # obj.m_w = parent.m_w @ obj.m_p_i @ obj.m_b
    # obj.new_mw = m_diff @ obj.m_b  # if no parent

    # if no Parent:
    #     obj.new_mw = m_diff @ I @ I @ obj.m_b
    # else: #! we skip - since transforming parent is enough - child will inhering diff_mat transformation
        #? search for m_diff2 - that will give same result for new.m_w with, and without parent
        # parent.m_w @ obj.m_p_i @ m_diff2 @ obj.m_b  == m_diff @ obj.m_b
        # parent.m_w @ obj.m_p_i @ m_diff2 == m_diff
        # m_diff2 == obj.m_p_i.inv() @ parent.m_w.inv() @ m_diff



def apply_col_mat_inv_transform(obj, mat):
    ''' if col.inner_objs had mat applied, then apply mat.inverted() to empties to compensate movement'''
    for child in obj.children: #makes children stay in place
        child.matrix_parent_inverse = mat @ child.matrix_parent_inverse

    # if not obj.parent:
    obj.matrix_basis = obj.matrix_basis @ mat.inverted()  # rottate oposite dir the obj to the mesh
    # else:
    #     diff_mat2 = obj.matrix_basis.inverted() @ obj.matrix_parent_inverse.inverted() @ obj.parent.matrix_world.inverted() @obj.matrix_basis @ mat.inverted()
    #     obj.matrix_basis = obj.matrix_basis @ diff_mat2

    # diff_mat = mat.inverted()
    # #wha kind of diff_mat2 we need?
    # parent.m_w @ obj.m_p_i @ obj.m_b @ diff_mat2 == obj.m_b @ diff_mat
    # diff_mat2 == obj.m_b.inv() @ obj.m_p_i.inv() @ parent.m_w.inv() @obj.m_b @ diff_mat
    # diff_mat2 == obj.m_b.inv() @ obj.m_p_i.inv() @ parent.m_w.inv() @obj.m_b @ mat.inverted()



def matrix_decompose(matrix_world):
    ''' returns active_obj_mat_loc, active_obj_mat_rot, active_obj_mat_sca
        reconstruct by loc @ rotQuat @ scale '''

    loc, rotQuat, scale = matrix_world.decompose()

    active_obj_mat_loc = Matrix.Translation(loc)
    active_obj_mat_rot = rotQuat.to_matrix().to_4x4()
    active_obj_mat_sca = Matrix()
    for i in range(3):
        active_obj_mat_sca[i][i] = scale[i]

    return active_obj_mat_loc, active_obj_mat_rot, active_obj_mat_sca

def get_scale_mat(vec_scale):
    mat_scale = Matrix.Identity(4)
    mat_scale[0][0] = vec_scale.x
    mat_scale[1][1] = vec_scale.y
    mat_scale[2][2] = vec_scale.z
    return mat_scale


def get_mat_world_obj(obj):
    obj.matrix_basis == (Matrix.Translation(obj.location) @ obj.rotation_quaternion.to_matrix().to_4x4() @ get_scale_mat(obj.scale))
    # replace rotation_quaternion with rotation_euler if you are using Euler mode
    matrix_local = obj.matrix_parent_inverse @ obj.matrix_basis
    matrix_world = matrix_local if obj.parent is None else obj.parent.matrix_word @ matrix_local
    return matrix_world

def create_empty(name, matrix, display_type='PLAIN_AXES'):
    empty = bpy.data.objects.new(name, None)
    empty.matrix_world = matrix
    empty.empty_display_type = display_type
    return empty


def get_new_mat_parent_inv(old_parent_m_w, new_parent_m_w, old_obj_m_p_inv):
    ''' cos: old_m_w == new_m_w'''
    ''' so:  old_parent_m_w @ obj.m_p_inv1 @ obj.m_basis  ==  new_parent_m_w @ obj.m_p_inv2 @ obj.m_basis'''
    ''' so:  new_parent_m_w.inv() @ old_parent_m_w @ obj.m_p_inv1   ==   obj.m_p_inv2'''
    return new_parent_m_w.inverted() @ old_parent_m_w @ old_obj_m_p_inv
