# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from mathutils import Vector, Matrix
from ..helpers_group import getEditedBBoxVerts, get_addon_preferences, get_edited_colls
import blf
import gpu
from gpu_extras.batch import batch_for_shader
from bpy.app.handlers import persistent

EDIT_BBOX_VERTS = None # hold edited group bbox verts

# HANDLERS ###############################
@persistent
def gpro_scene_update(scene, depsgraph):
    check_obj_updated(depsgraph)

@persistent
def gpro_undo_handler(scene, depsgraph):
    global EDIT_BBOX_VERTS
    EDIT_BBOX_VERTS = None # clear edited bbox verts on file load

def gpro_clear_cache():
    global EDIT_BBOX_VERTS
    EDIT_BBOX_VERTS = None # clear edited bbox verts on file load

@persistent
def gpro_redo_handler(scene, depsgraph):
    global EDIT_BBOX_VERTS
    EDIT_BBOX_VERTS = None # clear edited bbox verts on file load

@persistent
def handler_reset_edit_bbox(scene, depsgraph):
    global EDIT_BBOX_VERTS
    EDIT_BBOX_VERTS = None # clear edited bbox verts on file load


def check_obj_updated(depsgraph):
    active_obj = bpy.context.active_object
    if not active_obj:
        return

    global EDIT_BBOX_VERTS
    if len(bpy.context.scene.storedGroupSettings) == 0:
        EDIT_BBOX_VERTS = None # clear edited bbox verts
        return

    if bpy.context.mode == "OBJECT" and hasattr(depsgraph, 'updates'): # only in object mode - so that it won't slow down edit mode..
        for update in depsgraph.updates:
            # print((f'ID {update.id}, geom {update.is_updated_geometry}, transform: {update.is_updated_transform}, is_eval: {update.id.is_evaluated}'))
            # geo_update=update.is_updated_geometry
            if (type(update.id) == bpy.types.Object and (update.is_updated_geometry or update.is_updated_transform)) \
                or type(update.id) == bpy.types.Scene: # detects if obj was added or removed in scene
                # print('Updating bbox verts')
                EDIT_BBOX_VERTS = getEditedBBoxVerts(bpy.context)

            return

# HANDLERS ###############################


##########################         Drawing UI and BOX openGl  stuff                ###############################################

def drawText():
    if not bpy.context.space_data.overlay.show_overlays:
        return
    storageList = bpy.context.scene.storedGroupSettings  # should be active but how get context...
    lengthStor = len(storageList)
    if lengthStor > 0:
        GroupNames = ''
        i = 0
        for storage in storageList:  # bread crumbs
            empty = bpy.data.objects.get(storage.currentEmptyName)
            if empty:
                edited_group = empty.instance_collection
                name = edited_group.name if edited_group else 'BrokenGroup'
                GroupNames = GroupNames + name
            i += 1
            if i != lengthStor:
                GroupNames += ' > '
        # glColor4f(1.0, 1.0, 1.0, 1.0)
        font_id = 0
        blf.position(0, 70, 30, 0)
        blf.size(0, 14, 72) if bpy.app.version < (4, 0, 0) else blf.size(0, 14)
        blf.draw(font_id, 'Editing(' + str(lengthStor) + '): ' + GroupNames)
    # restore defaults

bbEdges = [  # edges fo bbox - first bottm, then top, then sides
        [0, 1],
        [1, 2],
        [2, 3],
        [3, 0],
        [4, 5],
        [5, 6],
        [6, 7],
        [7, 4],
        [0, 4],
        [1, 5],
        [2, 6],
        [3, 7]
    ]

shader = gpu.shader.from_builtin('UNIFORM_COLOR') # was 3D_UNIFORM_COLOR

def get_all_scenes_edited_colls():
    all_edited_cols = {} # {storedGroupSettings: [edited_col]}
    for scn in bpy.data.scenes:
        storageList = scn.storedGroupSettings
        if len(storageList):
            edited_cols = {bpy.data.objects[storage.currentEmptyName].instance_collection.name:storage for storage in storageList if bpy.data.objects.get(storage.currentEmptyName)}
            all_edited_cols.update(edited_cols)
    return all_edited_cols



def DrawDynamicBBox():
    context = bpy.context
    if not context.space_data.overlay.show_overlays:
        return
    storageList = context.scene.storedGroupSettings
    addon_prefs = get_addon_preferences()
    edited_groups = get_edited_colls()
    settings = context.preferences.themes[0].view_3d
    edit_gr_idx = {coll.name:i for i,coll in enumerate(edited_groups) if coll}

    #* Drawing active BBox
    if edited_groups:
        edited_lib_groups = []
        for lib in bpy.data.libraries:
            edited_lib_groups.extend(get_edited_colls(lib))

        # edited_empty = bpy.data.objects[storageList[-1].currentEmptyName]
        # edit_bbox_verts = getEditedBBoxVerts(edited_empty)
        global EDIT_BBOX_VERTS
        if not EDIT_BBOX_VERTS:
            EDIT_BBOX_VERTS = getEditedBBoxVerts(context)
        if EDIT_BBOX_VERTS:
            gpu.state.depth_test_set("NONE")
            gpu.state.blend_set("ALPHA")
            gpu.state.line_width_set(addon_prefs.box_line_width)

            batch = batch_for_shader(shader, 'LINES', {"pos": EDIT_BBOX_VERTS})
            shader.bind()

            active_col = addon_prefs.edited_lib_box_color[:] if edited_groups[-1] in edited_lib_groups else addon_prefs.edited_box_color[:]
            shader.uniform_float("color", active_col)
            batch.draw(shader)


    #* Drawing other BBoxes
    if addon_prefs.groupBBox_use:  # only doo this for active object then break
        obj_list = [obj for obj in context.visible_objects if obj.type == 'EMPTY' and obj.instance_collection]
    else:
        obj_list = [obj for obj in context.selected_objects if obj.type == 'EMPTY' and obj.instance_collection]

    all_edited_groups = get_all_scenes_edited_colls() #  scene.storedGroupSettings: [edited.colls]
    gpu.state.line_width_set(1)
    gpu.state.blend_set("NONE")
    gpu.state.depth_test_set("LESS")

    bbox_col = addon_prefs.bbox_color[:]
    for obj in obj_list:
        bbox_co = list(obj.instance_collection.BBoxVertsList)
        # index = edit_gr_idx[obj.instance_collection.name] if obj.instance_collection.name in edit_gr_idx else -1
        edit_storage = all_edited_groups.get(obj.instance_collection.name)

        if edit_storage: #fix bbox on edited groups
            fixEditedGroupMatrix = bpy.data.objects[edit_storage.currentEmptyName].matrix_world.copy()
            fixEditedGroupMatrix.translation = Vector((0, 0, 0)) #for coll_offset
            bboxVerts = [obj.matrix_world @ fixEditedGroupMatrix @ Vector((bbox_co[i * 3], bbox_co[i * 3 + 1], bbox_co[i * 3 + 2])) for i in range(8)]  # for box vert list  3*8   (x,y,z)*8
        else:
            bboxVerts = [obj.matrix_world @ Vector((bbox_co[i * 3], bbox_co[i * 3 + 1], bbox_co[i * 3 + 2])) for i in range(8)]  # for box vert list  3*8   (x,y,z)*8


        edit_bbox_verts = [tuple(bboxVerts[vert]) for edge in bbEdges for vert in edge]
        batch = batch_for_shader(shader, 'LINES', {"pos": edit_bbox_verts})
        shader.bind()
        shader.uniform_float("color", (bbox_col[0], bbox_col[1], bbox_col[2], 1))
        if obj.select_get():
            shader.uniform_float("color", (settings.object_selected[0], settings.object_selected[1], settings.object_selected[2], 1))
            if context.active_object == obj:
                shader.uniform_float("color", (settings.object_active[0], settings.object_active[1], settings.object_active[2], 1))
        batch.draw(shader)


    # restore defaults
    gpu.state.line_width_set(1)
    gpu.state.depth_test_set("LESS")

    return
